(*
Author:  Akihisa Yamada (2018-2019)
License: LGPL (see file COPYING.LESSER)
*)
theory Complete_Relations
  imports HOL.Real Binary_Relations
begin

no_notation conversep ("(_\<inverse>\<inverse>)" [1000] 1000)

section {* Completeness of Relations *}


subsection {* Completeness *}

locale bounded = less_eq_syntax + assumes bounded: "\<exists>t. \<forall>x. x \<sqsubseteq> t"
begin

lemma ex_bound[intro!]: "Ex (bound (\<sqsubseteq>) X)" using bounded by auto

lemma ex_extreme_UNIV[intro!]: "Ex (extreme (\<sqsubseteq>) UNIV)" using bounded by auto

lemma dual_empty_complete[intro!]: "Ex (extreme_bound (\<sqsubseteq>)\<inverse>\<inverse> {})" by auto

lemma UNIV_complete[intro!]: "Ex (extreme_bound (\<sqsubseteq>) UNIV)" using bounded by blast

end

context
  fixes r :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<sqsubseteq>" 50)
begin

lemma bounded_iff_extreme_UNIV: "bounded (\<sqsubseteq>) \<longleftrightarrow> Ex (extreme (\<sqsubseteq>) UNIV)"
  by (auto simp:bounded_def)

lemma bounded_iff_UNIV_complete: "bounded (\<sqsubseteq>) \<longleftrightarrow> Ex (extreme_bound (\<sqsubseteq>) UNIV)"
  by (unfold bounded_def, blast)

lemma bounded_iff_dual_empty_complete: "bounded (\<sqsubseteq>) \<longleftrightarrow> Ex (extreme_bound (\<sqsubseteq>)\<inverse>\<inverse> {})"
  by (auto simp:bounded_def)

end


text {*
  The following assumes every pair has a supremum. I.e., a semilattice if underlying
  relation is an order.
*}

locale pair_complete = less_eq_syntax +
  assumes pair_complete: "Ex (extreme_bound (\<sqsubseteq>) {x,y})"
begin

lemma directed_UNIV[intro!]: "directed (\<sqsubseteq>) UNIV"
proof
  fix x y :: 'a
  from pair_complete[of x y] show "\<exists>z \<in> UNIV. x \<sqsubseteq> z \<and> y \<sqsubseteq> z" by auto
qed

end

sublocale total \<subseteq> pair_complete
proof (unfold_locales)
  fix x y
  show "Ex (extreme_bound (\<sqsubseteq>) {x, y})" by (cases x y rule:total_cases, auto)
qed

text {*
  The next one assumes every nonempty finite set has a supremum.
*}
locale finite_complete = less_eq_syntax +
  assumes finite_nonempty_complete: "finite X \<Longrightarrow> X \<noteq> {} \<Longrightarrow> Ex (extreme_bound (\<sqsubseteq>) X)"

sublocale finite_complete \<subseteq> pair_complete
  by (unfold_locales, intro finite_nonempty_complete, auto)

text {*
  The next one assumes every nonempty bounded set has a supremum.
*}
locale conditionally_complete = less_eq_syntax +
  assumes bounded_nonempty_complete:
  "Ex (bound (\<sqsubseteq>) X) \<Longrightarrow> X \<noteq> {} \<Longrightarrow> Ex (extreme_bound (\<sqsubseteq>) X)"
begin

lemma bounded_nonemptyE[elim!]:
  assumes "Ex (bound (\<sqsubseteq>) X)" and "X \<noteq> {}" and "Ex (extreme_bound (\<sqsubseteq>) X) \<Longrightarrow> X \<noteq> {} \<Longrightarrow> thesis"
  shows thesis
  using assms bounded_nonempty_complete by auto

lemma nonempty_imp_complete_iff_bounded:
  assumes "X \<noteq> {}" shows "Ex (extreme_bound (\<sqsubseteq>) X) \<longleftrightarrow> Ex (bound (\<sqsubseteq>) X)"
  using assms by (auto intro: bounded_nonempty_complete)

end

text {*
  Other well-known assumptions follow.
*}

locale omega_complete = less_eq_syntax +
  assumes monotone_seq_complete:
    "\<And>f :: nat \<Rightarrow> 'a. monotone (\<le>) (\<sqsubseteq>) f \<Longrightarrow> Ex (extreme_bound (\<sqsubseteq>) (range f))"

locale chain_complete = less_eq_syntax +
  assumes chain_nonempty_complete: "chain (\<sqsubseteq>) X \<Longrightarrow> X \<noteq> {} \<Longrightarrow> Ex (extreme_bound (\<sqsubseteq>) X)"
begin

lemma monotone_chain_complete:
  assumes C0: "C \<noteq> {}" and chain: "chain r C" and mono: "monotone r (\<sqsubseteq>) f"
  shows "Ex (extreme_bound (\<sqsubseteq>) (f`C))"
  apply (rule chain_nonempty_complete[OF monotone_chain_image[OF mono chain]])
  using C0 by auto

end

sublocale chain_complete \<subseteq> omega_complete
  by (unfold_locales, rule monotone_chain_complete, auto intro:chainI)

locale directed_complete = less_eq_syntax +
  assumes directed_nonempty_complete: "directed (\<sqsubseteq>) X \<Longrightarrow> X \<noteq> {} \<Longrightarrow> Ex (extreme_bound (\<sqsubseteq>) X)"
begin

lemma monotone_directed_complete:
  assumes dir: "directed r C" and c0: "C \<noteq> {}" and mono: "monotone r (\<sqsubseteq>) f"
  shows "Ex (extreme_bound (\<sqsubseteq>) (f`C))"
  apply (rule directed_nonempty_complete[OF monotone_directed_image[OF mono dir]])
  using c0 by auto

end

sublocale directed_complete \<subseteq> chain_complete
  by (unfold_locales, intro directed_nonempty_complete, auto dest: chain_imp_directed)

text {*
  The next one is quite complete, only the empty set may fail to have a supremum.
*}
locale nonempty_complete = less_eq_syntax +
  assumes nonempty_complete: "X \<noteq> {} \<Longrightarrow> Ex (extreme_bound (\<sqsubseteq>) X)"

sublocale nonempty_complete \<subseteq>
  conditionally_complete + finite_complete + directed_complete
  by (unfold_locales, auto intro!: nonempty_complete)

sublocale nonempty_complete \<subseteq> bounded
  unfolding bounded_iff_UNIV_complete using nonempty_complete[of UNIV] by auto

subsubsection {* Pointed Ones *}

text {*
  The term `pointed' refers to the dual notion of boundedness, i.e., there is a global least element.
  This serves as the supremum of the empty set.
*}

text {*
 ``Bounded complete'' refers to pointed conditional complete, but this notion is just a dual of
  nonempty completeness. We prove this later.
 *}

locale pointed_chain_complete = less_eq_syntax +
  assumes chain_complete: "chain (\<sqsubseteq>) X \<Longrightarrow> Ex (extreme_bound (\<sqsubseteq>) X)"
begin

sublocale chain_complete using chain_complete by unfold_locales

sublocale dual: bounded "(\<sqsubseteq>)\<inverse>\<inverse>" using chain_complete[of "{}"]
  by (auto simp: bounded_iff_dual_empty_complete chain_empty)

end

locale pointed_directed_complete = less_eq_syntax +
  assumes directed_complete: "directed (\<sqsubseteq>) X \<Longrightarrow> Ex (extreme_bound (\<sqsubseteq>) X)"
begin

sublocale directed_complete using directed_complete by unfold_locales

sublocale dual: bounded "(\<sqsubseteq>)\<inverse>\<inverse>" using directed_complete[of "{}"]
  by (auto simp: bounded_iff_dual_empty_complete)

sublocale pointed_chain_complete
proof unfold_locales
  fix X
  assume "chain (\<sqsubseteq>) X"
  from chain_nonempty_complete[OF this]
  show "Ex (extreme_bound (\<sqsubseteq>) X)" by auto
qed

end

locale complete = less_eq_syntax +
  assumes complete: "Ex (extreme_bound (\<sqsubseteq>) X)"
begin

sublocale nonempty_complete by (unfold_locales, auto intro!: complete)

sublocale dual: bounded "(\<sqsubseteq>)\<inverse>\<inverse>" using complete[of "{}"] by (unfold_locales, auto)

end



subsubsection{* Relations between Completeness Conditions *}

context
  fixes less_eq :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<sqsubseteq>" 50)
begin

interpretation less_eq_dualize.

text {* Pair-completeness implies the universe is directed. *}

proposition directed_complete_pair_complete_imp_bounded:
  assumes "directed_complete (\<sqsubseteq>)" and "pair_complete (\<sqsubseteq>)"
  shows "bounded (\<sqsubseteq>)"
proof-
  from assms interpret directed_complete + pair_complete by auto
  have "Ex (extreme_bound (\<sqsubseteq>) UNIV)" by (rule directed_nonempty_complete, auto)
  then obtain t where "extreme_bound (\<sqsubseteq>) UNIV t" by auto
  then have "\<forall>x. x \<sqsubseteq> t" by auto
  then show ?thesis by (unfold_locales, auto)
qed

text {* The empty set has supremum iff there is a bottom element. *}

proposition complete_iff_nonempty_complete_pointed:
  "complete (\<sqsubseteq>) \<longleftrightarrow> nonempty_complete (\<sqsubseteq>) \<and> bounded (\<sqsupseteq>)"
proof (safe)
  assume "bounded (\<sqsupseteq>)" and "nonempty_complete (\<sqsubseteq>)"
  then interpret nonempty_complete "(\<sqsubseteq>)" + dual: bounded "(\<sqsupseteq>)" by auto
  show "complete (\<sqsubseteq>)"
    using dual.bounded nonempty_complete by (unfold_locales, auto)
next
  assume "complete (\<sqsubseteq>)"
  then interpret complete "(\<sqsubseteq>)".
  show "bounded (\<sqsupseteq>)" using dual.bounded_axioms.
  show "nonempty_complete (\<sqsubseteq>)"..
qed

text {*
  If the universe is bounded (topped), then every nonempty set has a supremum.
*}
proposition nonempty_complete_iff_conditionally_complete_bounded:
  "nonempty_complete (\<sqsubseteq>) \<longleftrightarrow> conditionally_complete (\<sqsubseteq>) \<and> bounded (\<sqsubseteq>)"
proof safe
  assume "conditionally_complete (\<sqsubseteq>)" and "bounded (\<sqsubseteq>)"
  then interpret conditionally_complete "(\<sqsubseteq>)" + bounded "(\<sqsubseteq>)".
  show "nonempty_complete (\<sqsubseteq>)"
    by (unfold_locales, rule bounded_nonempty_complete, auto)
next
  assume "nonempty_complete (\<sqsubseteq>)"
  then interpret nonempty_complete.
  show "bounded (\<sqsubseteq>)"..
  show "conditionally_complete (\<sqsubseteq>)"..
qed

text {*
  Conditional completeness only lacks top and bottom to be complete.
*}
proposition complete_iff_conditionally_complete_bounded_pointed:
  "complete (\<sqsubseteq>) \<longleftrightarrow> conditionally_complete (\<sqsubseteq>) \<and> bounded (\<sqsubseteq>) \<and> bounded (\<sqsupseteq>)"
  by (auto simp: complete_iff_nonempty_complete_pointed nonempty_complete_iff_conditionally_complete_bounded)

end


text {*
  If the universe is directed, then every pair is bounded, and thus has a supremum.
  On the other hand, supremum gives an upper bound, witnessing directedness.
*}
proposition(in conditionally_complete) pair_complete_iff_directed:
  "pair_complete (\<sqsubseteq>) \<longleftrightarrow> directed (\<sqsubseteq>) UNIV"
proof(intro iffI)
  assume "directed (\<sqsubseteq>) UNIV"
  then show "pair_complete (\<sqsubseteq>)"
    by (unfold_locales, intro bounded_nonempty_complete, auto elim: directedE)
next
  assume "pair_complete (\<sqsubseteq>)"
  then interpret pair_complete.
  show "directed (\<sqsubseteq>) UNIV"
  proof (intro directedI)
    fix x y
    from pair_complete obtain z where "extreme_bound (\<sqsubseteq>) {x,y} z" by auto
    then show "\<exists>z\<in>UNIV. x \<sqsubseteq> z \<and> y \<sqsubseteq> z" by auto
  qed
qed


subsubsection {* Completeness Results Requiring Order-Like Properties *}

text {*
  Above results hold without any assumption on the relation. This part demands
  some order-like properties.
*}

text {*
  Following is a folklore, semilattice is finite complete. This fact demands only
  transitivity of the underlying relation.
*}

locale trans_semilattice = transitive + pair_complete

sublocale trans_semilattice \<subseteq> finite_complete
  apply (unfold_locales)
  subgoal for X
  proof (induct X rule:finite_induct)
    case empty
    then show ?case by auto
  next
    case (insert x X)
    show ?case
    proof (cases "X = {}")
      case True
      obtain x' where "extreme_bound (\<sqsubseteq>) {x,x} x'" using pair_complete[of x x] by auto
      with True show ?thesis by (auto intro!: exI[of _ x'])
    next
      case False
      with insert obtain b where b: "extreme_bound (\<sqsubseteq>) X b" by auto
      from pair_complete obtain c where c: "extreme_bound (\<sqsubseteq>) {x,b} c" by auto
      show ?thesis
      proof (intro exI extreme_boundI)
        from c have "x \<sqsubseteq> c" and "b \<sqsubseteq> c" by auto
        with b show "xb \<in> insert x X \<Longrightarrow> xb \<sqsubseteq> c" for xb by (auto dest: trans)
        fix d assume "bound (\<sqsubseteq>) (insert x X) d"
        with b have "bound (\<sqsubseteq>) {x,b} d" by auto
        with c show "c \<sqsubseteq> d" by auto
      qed
    qed
  qed
  done


text {*
  Directed complete order is nonempty complete iff it is also a semilattice.
*}
proposition(in transitive) nonempty_complete_iff_directed_complete_pair_complete:
  "nonempty_complete (\<sqsubseteq>) \<longleftrightarrow> directed_complete (\<sqsubseteq>) \<and> pair_complete (\<sqsubseteq>)"
proof (safe, unfold_locales)
  assume "directed_complete (\<sqsubseteq>)" and "pair_complete (\<sqsubseteq>)"
  then interpret directed_complete + pair_complete.
  interpret trans_semilattice..
  fix X :: "'a set"
  have 1: "directed (\<sqsubseteq>) {x. \<exists>Y \<subseteq> X. finite Y \<and> Y \<noteq> {} \<and> extreme_bound (\<sqsubseteq>) Y x}" (is "directed _ ?B")
  proof (intro directedI)
    fix a b assume a: "a \<in> ?B" and b: "b \<in> ?B"
    from a obtain A where A: "extreme_bound (\<sqsubseteq>) A a" "finite A" "A \<noteq> {}" "A \<subseteq> X" by auto
    from b obtain B where B: "extreme_bound (\<sqsubseteq>) B b" "finite B" "B \<noteq> {}" "B \<subseteq> X" by auto
    from A B have AB: "finite (A \<union> B)" "A \<union> B \<noteq> {}" "A \<union> B \<subseteq> X" by auto
    with finite_nonempty_complete have "Ex (extreme_bound (\<sqsubseteq>) (A \<union> B))" by auto
    then obtain c where c: "extreme_bound (\<sqsubseteq>) (A \<union> B) c" by auto
    show "\<exists>c \<in> ?B. a \<sqsubseteq> c \<and> b \<sqsubseteq> c"
    proof (intro bexI conjI)
      from A B c show "a \<sqsubseteq> c" and "b \<sqsubseteq> c" by (auto simp: extreme_bound_iff)
      from AB c show "c \<in> ?B" by (auto intro!: exI[of _ "A \<union> B"])
    qed
  qed
  assume "X \<noteq> {}"
  then obtain x where xX: "x \<in> X" by auto
  from finite_nonempty_complete[of "{x}"]
  obtain x' where "extreme_bound (\<sqsubseteq>) {x} x'" by auto
  with xX have x'B: "x' \<in> ?B" by (auto intro!: exI[of _ "{x}"] extreme_boundI)
  then have 2: "?B \<noteq> {}" by auto
  from directed_nonempty_complete[OF 1 2]
  obtain b where b: "extreme_bound (\<sqsubseteq>) ?B b" by auto
  show "Ex (extreme_bound (\<sqsubseteq>) X)"
  proof (intro exI extreme_boundI UNIV_I)
    fix x
    assume xX: "x \<in> X"
    from finite_nonempty_complete[of "{x}"]
    obtain c where c: "extreme_bound (\<sqsubseteq>) {x} c" by auto
    then have xc: "x \<sqsubseteq> c" by auto
    from c xX have cB: "c \<in> ?B" by (auto intro!: exI[of _ "{x}"] extreme_boundI)
    with b have cb: "c \<sqsubseteq> b" by auto
    from xc cb show "x \<sqsubseteq> b" by (rule trans) text{* Here transitivity is needed. *}
  next
    fix x
    assume Xx: "bound (\<sqsubseteq>) X x"
    have "bound (\<sqsubseteq>) ?B x"
    proof (intro boundI UNIV_I, clarify)
      fix c Y
      assume "finite Y" and YX: "Y \<subseteq> X" and "Y \<noteq> {}" and c: "extreme_bound (\<sqsubseteq>) Y c"
      from YX Xx have "bound (\<sqsubseteq>) Y x" by auto
      with c show "c \<sqsubseteq> x" by auto
    qed
    with b show "b \<sqsubseteq> x" by auto
  qed
next
  assume "nonempty_complete (\<sqsubseteq>)"
  then interpret nonempty_complete.
  show "Ex (extreme_bound (\<sqsubseteq>) {x,y})" for x y by (rule nonempty_complete, auto)
  show "X \<noteq> {} \<Longrightarrow> Ex (extreme_bound (\<sqsubseteq>) X)" for X by (rule nonempty_complete, auto)
qed

proposition (in reflexive) nonempty_complete_iff_directed_complete_finite_complete:
  "nonempty_complete (\<sqsubseteq>) \<longleftrightarrow> directed_complete (\<sqsubseteq>) \<and> finite_complete (\<sqsubseteq>)"
proof (safe, unfold_locales)
  assume "directed_complete (\<sqsubseteq>)" and "finite_complete (\<sqsubseteq>)"
  then interpret directed_complete + finite_complete.
  fix X :: "'a set"
  have 1: "directed (\<sqsubseteq>) {x. \<exists>Y \<subseteq> X. finite Y \<and> Y \<noteq> {} \<and> extreme_bound (\<sqsubseteq>) Y x}" (is "directed _ ?B")
  proof (intro directedI)
    fix a b assume a: "a \<in> ?B" and b: "b \<in> ?B"
    from a obtain A where A: "extreme_bound (\<sqsubseteq>) A a" "finite A" "A \<noteq> {}" "A \<subseteq> X" by auto
    from b obtain B where B: "extreme_bound (\<sqsubseteq>) B b" "finite B" "B \<noteq> {}" "B \<subseteq> X" by auto
    from A B have AB: "finite (A \<union> B)" "A \<union> B \<noteq> {}" "A \<union> B \<subseteq> X" by auto
    with finite_nonempty_complete have "Ex (extreme_bound (\<sqsubseteq>) (A \<union> B))" by auto
    then obtain c where c: "extreme_bound (\<sqsubseteq>) (A \<union> B) c" by auto
    show "\<exists>c \<in> ?B. a \<sqsubseteq> c \<and> b \<sqsubseteq> c"
    proof (intro bexI conjI)
      from A B c show "a \<sqsubseteq> c" and "b \<sqsubseteq> c" by (auto simp: extreme_bound_iff)
      from AB c show "c \<in> ?B" by (auto intro!: exI[of _ "A \<union> B"])
    qed
  qed
  assume "X \<noteq> {}"
  then obtain x where xX: "x \<in> X" by auto
  then have "extreme_bound (\<sqsubseteq>) {x} x" by auto
  with xX have xB: "x \<in> ?B" by (auto intro!: exI[of _ "{x}"] extreme_boundI)
  then have 2: "?B \<noteq> {}" by auto
  from directed_nonempty_complete[OF 1 2]
  obtain b where b: "extreme_bound (\<sqsubseteq>) ?B b" by auto
  show "Ex (extreme_bound (\<sqsubseteq>) X)"
  proof (intro exI extreme_boundI UNIV_I)
    fix x
    assume xX: "x \<in> X"
    have x: "extreme_bound (\<sqsubseteq>) {x} x" by auto
    from x xX have cB: "x \<in> ?B" by (auto intro!: exI[of _ "{x}"] extreme_boundI)
    with b show "x \<sqsubseteq> b" by auto
  next
    fix x
    assume Xx: "bound (\<sqsubseteq>) X x"
    have "bound (\<sqsubseteq>) ?B x"
    proof (intro boundI UNIV_I, clarify)
      fix c Y
      assume "finite Y" and YX: "Y \<subseteq> X" and "Y \<noteq> {}" and c: "extreme_bound (\<sqsubseteq>) Y c"
      from YX Xx have "bound (\<sqsubseteq>) Y x" by auto
      with c show "c \<sqsubseteq> x" by auto
    qed
    with b show "b \<sqsubseteq> x" by auto
  qed
next
  assume "nonempty_complete (\<sqsubseteq>)"
  then interpret nonempty_complete.
  fix X :: "'a set" assume "X \<noteq> {}"
  then show "Ex (extreme_bound (\<sqsubseteq>) X)" by (intro nonempty_complete, auto)
  then show "Ex (extreme_bound (\<sqsubseteq>) X)".
qed

locale conditionally_complete_pseudo_order = pseudo_order + conditionally_complete

locale conditionally_complete_partial_order = partial_order + conditionally_complete
begin

sublocale conditionally_complete_pseudo_order..

end

locale complete_trellis = complete + pseudo_order
begin

sublocale conditionally_complete_pseudo_order..

end

text {* Finally, we (re)define complete lattices as complete partial order. *}
locale complete_lattice_locale = partial_order + complete
begin

sublocale trans_semilattice..

sublocale conditionally_complete_partial_order..

sublocale complete_trellis..

end


subsection {* Relating to Classes *}

class conditionally_complete_psorder = psorder +
  assumes "conditionally_complete (\<le>)"
begin

sublocale ord: conditionally_complete
  using conditionally_complete_psorder_axioms
  unfolding class.conditionally_complete_psorder_def
  unfolding class.conditionally_complete_psorder_axioms_def
  by auto

sublocale ord: conditionally_complete_pseudo_order..

end

class conditionally_complete_worder = worder +
  assumes "conditionally_complete (\<le>)"
begin

interpretation ord: conditionally_complete
  using conditionally_complete_worder_axioms
  unfolding class.conditionally_complete_worder_def
  unfolding class.conditionally_complete_worder_axioms_def
  by auto

subclass conditionally_complete_psorder..

sublocale ord: conditionally_complete_partial_order..

end

text {* Isabelle's class @{class Conditionally_Complete_Lattices.conditionally_complete_lattice}
  is @{class conditionally_complete_psorder}.
*}

subclass (in Conditionally_Complete_Lattices.conditionally_complete_lattice) conditionally_complete_psorder
proof
  fix X
  assume "Ex (upper_bound X)" and X0: "X \<noteq> {}"
  from this(1) have "bdd_above X" by auto
  from cSup_upper[OF _ this] cSup_least[OF X0]
  have "least {b. upper_bound X b} (Sup X)" by (intro extremeI boundI, auto)
  then show "Ex (least {b. upper_bound X b})" by auto
qed



context complete_lattice begin

interpretation ord: complete_lattice_locale
  by (unfold_locales, auto intro!: extremeI boundI Sup_upper Sup_least Inf_lower Inf_greatest)

subclass conditionally_complete_worder..
sublocale ord: complete_lattice_locale..

end


subsubsection {* Duality of Completeness Conditions *}

text {* Conditional completeness is symmetric. *}

sublocale conditionally_complete \<subseteq> dual: conditionally_complete "(\<sqsubseteq>)\<inverse>\<inverse>"
proof
  fix X :: "'a set"
  assume bound: "Ex (bound (\<sqsubseteq>)\<inverse>\<inverse> X)" and nonemp: "X \<noteq> {}"
  then have "Ex (bound (\<sqsubseteq>) {b. bound (\<sqsubseteq>)\<inverse>\<inverse> X b})" and "{b. bound (\<sqsubseteq>)\<inverse>\<inverse> X b} \<noteq> {}" by auto
  from bounded_nonempty_complete[OF this]
  obtain b where b: "extreme_bound (\<sqsubseteq>) {b. bound (\<sqsubseteq>)\<inverse>\<inverse> X b} b" by auto
  then show "Ex (extreme_bound (\<sqsubseteq>)\<inverse>\<inverse> X)" by (auto intro!: exI[of _ b] extremeI)
qed

context conditionally_complete begin
lemmas conditionally_complete_dual = dual.conditionally_complete_axioms
end

sublocale conditionally_complete_pseudo_order \<subseteq> dual: conditionally_complete_pseudo_order "(\<sqsubseteq>)\<inverse>\<inverse>"..

sublocale conditionally_complete_partial_order \<subseteq> dual: conditionally_complete_partial_order "(\<sqsubseteq>)\<inverse>\<inverse>"..


text {* Full completeness is symmetric. *}

sublocale complete \<subseteq> dual: complete "(\<sqsubseteq>)\<inverse>\<inverse>"
proof
  fix X :: "'a set"
  obtain s where "extreme_bound (\<sqsubseteq>) {b. bound (\<sqsubseteq>)\<inverse>\<inverse> X b} s" using complete by auto
  then show "Ex (extreme_bound (\<sqsubseteq>)\<inverse>\<inverse> X)" by (intro exI[of _ s] extreme_boundI, auto)
qed

sublocale complete_trellis \<subseteq> dual: complete_trellis "(\<sqsubseteq>)\<inverse>\<inverse>"..

sublocale complete_lattice_locale \<subseteq> dual: complete_lattice_locale "(\<sqsubseteq>)\<inverse>\<inverse>"..

text {* Now we show bounded completeness is dual of nonempty completeness.
*}

context fixes less_eq :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<sqsubseteq>" 50)
begin

interpretation less_eq_dualize.

definition "bounded_complete \<equiv> \<forall>X. Ex (bound (\<sqsubseteq>) X) \<longrightarrow> Ex (extreme_bound (\<sqsubseteq>) X)"

lemma pointed_conditionally_complete_iff_bounded_complete:
  "conditionally_complete (\<sqsubseteq>) \<and> bounded (\<sqsupseteq>) \<longleftrightarrow> bounded_complete"
proof safe
  assume "bounded_complete"
  note * = this[unfolded bounded_complete_def, rule_format]
  from * show "conditionally_complete (\<sqsubseteq>)" by (unfold_locales, auto)
  from *[of "{}"] show "bounded (\<sqsupseteq>)" by (unfold_locales, auto)
next
  assume "conditionally_complete (\<sqsubseteq>)" and "bounded (\<sqsupseteq>)"
  then interpret conditionally_complete "(\<sqsubseteq>)" + dual: bounded "(\<sqsupseteq>)".
  show "bounded_complete" unfolding bounded_complete_def
  proof (intro allI impI)
    fix X
    assume X: "Ex (bound (\<sqsubseteq>) X)"
    show "Ex (extreme_bound (\<sqsubseteq>) X)"
    proof (cases "X = {}")
    case True
    then show ?thesis by auto
    next
      case False
      with bounded_nonempty_complete X show ?thesis by auto
    qed
  qed
qed

proposition bounded_complete_iff_dual_nonempty_complete:
  "bounded_complete \<longleftrightarrow> nonempty_complete (\<sqsupseteq>)"
proof (fold pointed_conditionally_complete_iff_bounded_complete, safe)
  assume "conditionally_complete (\<sqsubseteq>)" and "bounded (\<sqsupseteq>)"
  then interpret conditionally_complete + bounded "(\<sqsupseteq>)".
  from dual.conditionally_complete_axioms bounded_axioms
    nonempty_complete_iff_conditionally_complete_bounded
  show "nonempty_complete (\<sqsupseteq>)" by auto
next
  assume "nonempty_complete (\<sqsupseteq>)"
  then interpret nonempty_complete "(\<sqsupseteq>)".
  show "conditionally_complete (\<sqsubseteq>)"..
  show "bounded (\<sqsupseteq>)"..
qed

end
  
subsection {* Interpretations *}

instance real :: conditionally_complete_worder
  by (intro_classes, unfold_locales)

instance "fun" :: (type, conditionally_complete_worder) conditionally_complete_worder
proof (intro_classes, unfold_locales)
  fix F :: "('a \<Rightarrow> 'b) set"
  let ?F = "\<lambda>x. {f x |. f \<in> F}"
  assume ex: "Ex (upper_bound F)" and F0: "F \<noteq> {}"
  { fix x
    from ex obtain b where "upper_bound F b" by auto
    then have "upper_bound (?F x) (b x)" by (auto simp: le_fun_def)
    then have "Ex (upper_bound (?F x))" by force
  }
  from ord.bounded_nonempty_complete[OF this] F0
  have "\<forall>x. Ex (least {b. upper_bound (?F x) b})" by auto
  from choice[OF this] fun_lub_iff
  show "Ex (least {b. upper_bound F b})" by auto
qed

end